"""Library to handle connection with UberSolar."""
from __future__ import annotations

from bleak_retry_connector import close_stale_connections, get_device

from .devices.ubersmart import UberSmart
from .discovery import GetUberSolarDevices
from .models import UberSolarAdvertisement

__all__ = [
    "get_device",
    "close_stale_connections",
    "GetUberSolarDevices",
    "UberSolarAdvertisement",
    "UberSmart",
]



"""UberSolar get device info test."""
import argparse
import asyncio
import logging

from . import GetUberSolarDevices, UberSmart

logger = logging.getLogger(__name__)


async def main(args: argparse.Namespace):
    """Main program."""

    logger.info("starting scan...")
    device = await GetUberSolarDevices().discover()

    if not device:
        logger.info("Could not find any UberSolar devices in range")
        return

    logger.info("connecting to device and getting info...")

    if args.address:
        uber_smart = UberSmart(device[args.address].device)
        await uber_smart.get_info()
        print(uber_smart.status_data[args.address])

        if args.switch:
            logger.info("setting switches...")
            await uber_smart.toggle_switches_all(args.switch)

        if args.wifiap:
            logger.info("enable ap sta on device...")
            await uber_smart.enable_wifi_ap()

        if args.time:
            logger.info("setting time on device...")
            await uber_smart.set_current_time()

    else:
        for item in device:
            uber_smart = UberSmart(device[item].device)
            await uber_smart.get_info()
            print(uber_smart.status_data)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--address",
        metavar="<address>",
        help="the address of the bluetooth device to connect to",
    )

    # elementToggle, pumpToggle, holidayModeToggle, solenoidMode = 0 - off, 1 - on, 2 - auto
    parser.add_argument(
        "--switch",
        metavar="<switch>",
        help="Set switches on device",
    )
    parser.add_argument(
        "--wifiap",
        action="store_true",
        help="enable wifiap on device",
    )
    parser.add_argument(
        "--time",
        action="store_true",
        help="Update device date and time",
    )

    parser.add_argument(
        "-d",
        "--debug",
        action="store_true",
        help="sets the log level to debug",
    )

    args = parser.parse_args()

    LOG_LEVEL = logging.DEBUG if args.debug else logging.INFO
    logging.basicConfig(
        level=LOG_LEVEL,
        format="%(asctime)-15s %(name)-8s %(levelname)s: %(message)s",
    )

    asyncio.run(main(args))



"""Discover UberSolar devices."""

from __future__ import annotations

import asyncio
import logging

import bleak

from .const import DEFAULT_RETRY_COUNT, DEFAULT_RETRY_TIMEOUT, DEFAULT_SCAN_TIMEOUT
from .models import UberSolarAdvertisement

_LOGGER = logging.getLogger(__name__)
CONNECT_LOCK = asyncio.Lock()


class GetUberSolarDevices:
    """Scan for all UberSolar devices."""

    def __init__(self, interface: int = 0) -> None:
        """Get UberSolar devices class constructor."""
        self._interface = f"hci{interface}"
        self._adv_data: dict[str, UberSolarAdvertisement] = {}

    def detection_callback(
        self,
        device: bleak.backends.device.BLEDevice,
        advertisement_data: bleak.backends.scanner.AdvertisementData,
    ) -> None:
        """BTLE adv scan callback."""

        if "UberSmart_" in device.name:
            discovery = UberSolarAdvertisement(
                device.address, device.name, device, advertisement_data.rssi, True
            )
            if discovery:
                self._adv_data[discovery.address] = discovery

    async def discover(
        self, retry: int = DEFAULT_RETRY_COUNT, scan_timeout: int = DEFAULT_SCAN_TIMEOUT
    ) -> dict:
        """Find UberSolar devices."""

        devices = None
        devices = bleak.BleakScanner(
            adapter=self._interface,
            detection_callback=self.detection_callback,
        )

        async with CONNECT_LOCK:
            await devices.start()
            await asyncio.sleep(scan_timeout)
            await devices.stop()

        if devices is None:
            if retry < 1:
                _LOGGER.error(
                    "Scanning for UberSolar devices failed. Stop trying", exc_info=True
                )
                return self._adv_data

            _LOGGER.warning(
                "Error scanning for UberSolar devices. Retrying (remaining: %d)",
                retry,
            )
            await asyncio.sleep(DEFAULT_RETRY_TIMEOUT)
            return await self.discover(retry - 1, scan_timeout)

        return self._adv_data

    async def get_device_data(
        self, address: str
    ) -> dict[str, UberSolarAdvertisement] | None:
        """Return data for specific device."""
        if not self._adv_data:
            await self.discover()

        return {
            device: adv
            for device, adv in self._adv_data.items()
            # MacOS uses UUIDs instead of MAC addresses
            if adv.data.get("address") == address
        }



"""Library to handle connection with UberSolar."""
from __future__ import annotations

import asyncio
from collections.abc import Callable
import logging
import time
from typing import Any
from uuid import UUID

from bleak import BleakError
from bleak.backends.device import BLEDevice
from bleak.backends.service import BleakGATTCharacteristic, BleakGATTServiceCollection
from bleak.exc import BleakDBusError
from bleak_retry_connector import (
    BLEAK_RETRY_EXCEPTIONS,
    BleakClientWithServiceCache,
    BleakNotFoundError,
    establish_connection,
)

from ..adv_parsers.ubersmart import process_ubersmart
from ..const import DEFAULT_RETRY_COUNT

_LOGGER = logging.getLogger(__name__)

# How long to hold the connection
# to wait for additional commands for
# disconnecting the device.
DISCONNECT_DELAY = 10

# We need to poll to get data
POLL_INTERVAL = 60


class CharacteristicMissingError(Exception):
    """Raised when a characteristic is missing."""


class UberSmartOperationError(Exception):
    """Raised when an operation fails."""


def _uuid(comms_type: str = "service") -> UUID | str:
    """Return UberSolar UUID."""

    _uuid = {
        "tx": "9af90c38",
        "rx": "95b10712",
        "service": "88665b98",
        "FWVersion": "9af90c40",
    }

    if comms_type in _uuid:
        return UUID(f"{_uuid[comms_type]}-0bfb-11ec-9a03-0242ac130003")

    return "Incorrect type, choose between: tx, rx, FWVersion or service"


READ_CHAR_UUID = _uuid(comms_type="tx")
WRITE_CHAR_UUID = _uuid(comms_type="rx")
SERVICE_CHAR_UUID = _uuid()
FW_CHAR_UUID = _uuid(comms_type="FWVersion")


class UberSolarBaseDevice:
    """Base Representation of a UberSolar Device."""

    def __init__(
        self,
        device: BLEDevice,
        interface: int = 0,
        **kwargs: Any,
    ) -> None:
        """Switchbot base class constructor."""
        self._interface = f"hci{interface}"
        self._device = device
        self._retry_count: int = kwargs.pop("retry_count", DEFAULT_RETRY_COUNT)
        self._connect_lock = asyncio.Lock()
        self._operation_lock = asyncio.Lock()
        self._client: BleakClientWithServiceCache | None = None
        self._read_char: BleakGATTCharacteristic | None = None
        self._write_char: BleakGATTCharacteristic | None = None
        self._disconnect_timer: asyncio.TimerHandle | None = None
        self._expected_disconnect = False
        self.loop = asyncio.get_event_loop()
        self._callbacks: list[Callable[[], None]] = []
        self._notify_future: asyncio.Future[bytearray] | None = None
        self.status_data: dict[str, Any] = {}
        self._last_full_update: float = -POLL_INTERVAL

    async def _send_command(
        self, key: str = "", retry: int | None = None
    ) -> bytes | None:
        """Send command to device and read response."""
        if retry is None:
            retry = self._retry_count

        command = bytearray.fromhex(key)
        _LOGGER.debug("%s: Scheduling command %s", self.name, key)
        max_attempts = retry + 1
        if self._operation_lock.locked():
            _LOGGER.debug(
                "%s: Operation already in progress, waiting for it to complete; RSSI: %s",
                self.name,
                self.rssi,
            )
        async with self._operation_lock:
            for attempt in range(max_attempts):
                try:
                    return await self._send_command_locked(command)
                except BleakNotFoundError:
                    _LOGGER.error(
                        "%s: device not found, no longer in range, or poor RSSI: %s",
                        self.name,
                        self.rssi,
                        exc_info=True,
                    )
                    raise
                except CharacteristicMissingError as ex:
                    if attempt == retry:
                        _LOGGER.error(
                            "%s: characteristic missing: %s; Stopping trying; RSSI: %s",
                            self.name,
                            ex,
                            self.rssi,
                            exc_info=True,
                        )
                        raise

                    _LOGGER.debug(
                        "%s: characteristic missing: %s; RSSI: %s",
                        self.name,
                        ex,
                        self.rssi,
                        exc_info=True,
                    )
                except BLEAK_RETRY_EXCEPTIONS:
                    if attempt == retry:
                        _LOGGER.error(
                            "%s: communication failed; Stopping trying; RSSI: %s",
                            self.name,
                            self.rssi,
                            exc_info=True,
                        )
                        raise

                    _LOGGER.debug(
                        "%s: communication failed with:", self.name, exc_info=True
                    )

        raise RuntimeError("Unreachable")

    @property
    def name(self) -> str:
        """Return device name."""
        return f"{self._device.name} ({self._device.address})"

    @property
    def rssi(self) -> int:
        """Return RSSI of device."""
        return self._device.rssi

    async def _ensure_connected(self) -> None:
        """Ensure connection to device is established."""
        if self._connect_lock.locked():
            _LOGGER.debug(
                "%s: Connection already in progress, waiting for it to complete; RSSI: %s",
                self.name,
                self.rssi,
            )
        if self._client and self._client.is_connected:
            self._reset_disconnect_timer()
            return
        async with self._connect_lock:
            # Check again while holding the lock
            if self._client and self._client.is_connected:
                self._reset_disconnect_timer()
                return
            _LOGGER.debug("%s: Connecting; RSSI: %s", self.name, self.rssi)
            client: BleakClientWithServiceCache = await establish_connection(
                BleakClientWithServiceCache,
                self._device,
                self.name,
                self._disconnected,
                use_services_cache=True,
                ble_device_callback=lambda: self._device,
            )
            _LOGGER.debug("%s: Connected; RSSI: %s", self.name, self.rssi)

            try:
                self._resolve_characteristics(client.services)
            except CharacteristicMissingError as ex:
                _LOGGER.debug(
                    "%s: characteristic missing, clearing cache: %s; RSSI: %s",
                    self.name,
                    ex,
                    self.rssi,
                    exc_info=True,
                )
                await client.clear_cache()
                await self._execute_forced_disconnect()
                raise

            self._client = client
            self._reset_disconnect_timer()
            await self._start_notify()

    def _resolve_characteristics(self, services: BleakGATTServiceCollection) -> None:
        """Resolve characteristics."""
        self._read_char = services.get_characteristic(READ_CHAR_UUID)
        if not self._read_char:
            raise CharacteristicMissingError(READ_CHAR_UUID)
        self._write_char = services.get_characteristic(WRITE_CHAR_UUID)
        if not self._write_char:
            raise CharacteristicMissingError(WRITE_CHAR_UUID)

    def _reset_disconnect_timer(self) -> None:
        """Reset disconnect timer."""
        self._cancel_disconnect_timer()
        self._expected_disconnect = False
        self._disconnect_timer = self.loop.call_later(
            DISCONNECT_DELAY, self._disconnect_from_timer
        )

    def _disconnected(self, client: BleakClientWithServiceCache) -> None:
        """Disconnected callback."""
        if self._expected_disconnect:
            _LOGGER.debug(
                "%s: Disconnected from device; RSSI: %s", self.name, self.rssi
            )
            return
        _LOGGER.warning(
            "%s: Device unexpectedly disconnected; RSSI: %s",
            self.name,
            self.rssi,
        )

    def _disconnect_from_timer(self) -> None:
        """Disconnect from device."""
        if self._operation_lock.locked() and self._client.is_connected:
            _LOGGER.debug(
                "%s: Operation in progress, resetting disconnect timer; RSSI: %s",
                self.name,
                self.rssi,
            )
            self._reset_disconnect_timer()
            return
        self._cancel_disconnect_timer()
        asyncio.create_task(self._execute_timed_disconnect())

    def _cancel_disconnect_timer(self) -> None:
        """Cancel disconnect timer."""
        if self._disconnect_timer:
            self._disconnect_timer.cancel()
            self._disconnect_timer = None

    async def _execute_forced_disconnect(self) -> None:
        """Execute forced disconnection."""
        self._cancel_disconnect_timer()
        _LOGGER.debug(
            "%s: Executing forced disconnect",
            self.name,
        )
        await self._execute_disconnect()

    async def _execute_timed_disconnect(self) -> None:
        """Execute timed disconnection."""
        _LOGGER.debug(
            "%s: Executing timed disconnect after timeout of %s",
            self.name,
            DISCONNECT_DELAY,
        )
        await self._execute_disconnect()

    async def _execute_disconnect(self) -> None:
        """Execute disconnection."""
        async with self._connect_lock:
            if self._disconnect_timer:  # If the timer was reset, don't disconnect
                return
            client = self._client
            self._expected_disconnect = True
            self._client = None
            self._read_char = None
            self._write_char = None
            if client and client.is_connected:
                _LOGGER.debug("%s: Disconnecting", self.name)
                await client.disconnect()
                _LOGGER.debug("%s: Disconnect completed", self.name)

    async def _send_command_locked(self, command: bytes) -> bytes:
        """Send command to device and read response."""
        await self._ensure_connected()
        if command != b"":
            try:
                return await self._execute_command_locked(command)
            except BleakDBusError as ex:
                # Disconnect so we can reset state and try again
                await asyncio.sleep(0.25)
                _LOGGER.debug(
                    "%s: RSSI: %s; Backing off %ss; Disconnecting due to error: %s",
                    self.name,
                    self.rssi,
                    0.25,
                    ex,
                )
                await self._execute_forced_disconnect()
                raise
            except BleakError as ex:
                # Disconnect so we can reset state and try again
                _LOGGER.debug(
                    "%s: RSSI: %s; Disconnecting due to error: %s",
                    self.name,
                    self.rssi,
                    ex,
                )
                await self._execute_forced_disconnect()
                raise

    def _notification_handler(self, _sender: int, data: bytearray) -> None:
        """Handle notification responses."""
        _LOGGER.debug("%s: Received notification: %s", self.name, data)
        self.status_data[self._device.address].update(process_ubersmart(data))

    async def _start_notify(self) -> None:
        """Start notification."""
        _LOGGER.debug("%s: Subscribe to notifications; RSSI: %s", self.name, self.rssi)

        self.status_data[self._device.address] = {}
        await self._client.start_notify(self._read_char, self._notification_handler)
        await asyncio.sleep(DISCONNECT_DELAY - 0.5)
        await self._client.stop_notify(self._read_char)

    async def _execute_command_locked(self, command: bytes) -> None:
        """Execute command and read response."""
        assert self._client is not None
        assert self._read_char is not None
        assert self._write_char is not None

        _LOGGER.debug("%s: Sending command: %s", self.name, command.hex())
        await self._client.write_gatt_char(self._write_char, command, False)

        return

    def get_address(self) -> str:
        """Return address of device."""
        return self._device.address

    def subscribe(self, callback: Callable[[], None]) -> Callable[[], None]:
        """Subscribe to device notifications."""
        self._callbacks.append(callback)

        def _unsub() -> None:
            """Unsubscribe from device notifications."""
            self._callbacks.remove(callback)

        return _unsub

    async def update(self) -> None:
        """Get device updates."""
        await self.get_info()
        self._last_full_update = time.monotonic()
        self._fire_callbacks()

    def _fire_callbacks(self) -> None:
        """Fire callbacks."""
        _LOGGER.debug("%s: Fire callbacks", self.name)
        for callback in self._callbacks:
            callback()

    async def get_info(self) -> dict[str, Any] | None:
        """Get device statuses."""

        await self._send_command()
        if not self.status_data[self._device.address]:
            _LOGGER.error("%s: Unsuccessful, no result from device", self.name)

        return self.status_data[self._device.address]

    def _check_command_result(
        self, result: bytes | None, index: int, values: set[int]
    ) -> bool:
        """Check command result."""
        if not result or len(result) - 1 < index:
            result_hex = result.hex() if result else "None"
            raise UberSmartOperationError(
                f"{self.name}: Sending command failed (result={result_hex} index={index} expected={values} rssi={self.rssi})"
            )
        return result[index] in values



"""Library to handle connection with UberSmart."""
from __future__ import annotations

import logging
import struct
import time
from typing import Any

from .device import UberSolarBaseDevice

_LOGGER = logging.getLogger(__name__)


class UberSmart(UberSolarBaseDevice):
    """Representation of a UberSmart Device."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """UberSolar UberSmart constructor."""

        super().__init__(*args, **kwargs)

    async def toggle_switches_all(self, switches: str) -> None:
        """Set pump switch."""

        # 1st byte is message id, use 06 to toggle switches.
        # need to send  all switches.
        # elementToggle, pumpToggle, holidayModeToggle, solenoidMode = 0 - off, 1 - on, 2 - auto
        # example 0600000002
        if not len(switches) == 8:
            _LOGGER.error("Switch length has to be 8")

        await self._send_command(key=f"06{switches}")
        _LOGGER.info("%s: Toggle switches", self.name)

    async def enable_wifi_ap(self) -> None:
        """Enable Wifi ap mode on device."""
        # intended for firmware update but who knows.
        await self._send_command(key="14")
        _LOGGER.info("%s: Enable Wifi AP on device", self.name)

    async def set_current_time(self) -> None:
        """Set current datetime on device."""

        current_time = int(time.time())
        ct_to_bytearray = bytearray(struct.pack("<Qxxxx", current_time))
        ct_to_bytearray[9] = 2  # add utc offset on byte 9

        await self._send_command(key=f"09{ct_to_bytearray.hex()}")
        _LOGGER.info("%s: Send current time to device", self.name)



"""Library to decode UberSmart Response."""
from __future__ import annotations

import datetime
import struct
from typing import Any


def process_ubersmart(data: bytearray) -> dict[str, Any]:
    """Process UberSmart data."""

    if data[0] == 1:
        return {
            "temps": {
                "fWaterTemperature": struct.unpack("<f", data[1:5])[0],
                "fManifoldTemperature": struct.unpack("<f", data[5:9])[0],
                "fStoredWater": struct.unpack("<f", data[9:13])[0],
            }
        }

    if data[0] == 2:
        return {
            "switches": {
                "bElementOn": data[1],
                "bPumpOn": data[2],
                "bHolidayMode": data[3],
                "eSolenoidMode": data[4],
                "fSolenoidState": struct.unpack("<f", data[5:9])[0],
            }
        }

    if data[0] == 3:
        _llu_time = struct.unpack("<Q", data[1:9])[0]
        return {
            "diag1": {
                "lluTime": datetime.datetime.fromtimestamp(_llu_time).strftime(
                    "%Y-%m-%d %H:%M:%S"
                ),
                "fHours": struct.unpack("<f", data[9:13])[0],  # Time on
                "wLux": struct.unpack("<H", data[13:15])[0],
            }
        }

    if data[0] == 4:
        return {
            "diag2": {
                "wRSSI": struct.unpack("<h", data[1:3])[0],
                "fPanelVoltage": struct.unpack("<f", data[3:7])[0],
                "fChipTemp": struct.unpack("<f", data[7:11])[0],
                "fWaterLevel": struct.unpack("<f", data[11:15])[0],
                "fTankSize": struct.unpack("<f", data[15:19])[0],
            }
        }

    if data[0] == 5:
        return {
            "FaultCodes": {
                "bPanelFaultCode": struct.unpack("<B", data[1:2])[0],
                "bElementFaultCode": struct.unpack("<B", data[2:3])[0],
                "bPumpFultCode": struct.unpack("<B", data[3:4])[0],
                "bSolenoidFaultCode": struct.unpack("<B", data[4:5])[0],
            }
        }

    return {}



"""Library to handle connection with UberSolar."""
from __future__ import annotations

from dataclasses import dataclass

from bleak.backends.device import BLEDevice


@dataclass
class UberSolarAdvertisement:
    """UberSolar advertisement."""

    address: str
    name: str
    device: BLEDevice
    rssi: int
    active: bool = False



from setuptools import setup

setup(
    name = "pyUbersolar",
    packages = ["ubersolar", "ubersolar.devices", "ubersolar.adv_parsers"],
    install_requires=["async_timeout>=4.0.1", "bleak>=0.17.0", "bleak-retry-connector>=2.9.0", "cryptography>=38.0.3"],
    version = "0.0.4",
    description = "A library to communicate with UberSolar devices",
    long_description= "API for accessing UberSolar UberSmart devices. Communication to/from the device via Bluetooth. Please view readme on github",
    author="Renier Moorcroft",
    url="https://github.com/RenierM26/pyUbersolar",
    license="MIT",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Home Automation",
        "Topic :: Software Development :: Libraries :: Python Modules"
        ],
)
